from qick.asm_v2 import AveragerProgramV2, QickSweep1D, AsmV2
from .helper import dB2gain
import numpy as np
import matplotlib.pyplot as plt
from scipy.ndimage import gaussian_filter

def generate_waveform(o, soccfg):
    f_fabric = soccfg["gens"][o["g"]]["f_fabric"]
    samps_per_clk = soccfg["gens"][o["g"]]["samps_per_clk"]
    l = int(o["length"] * f_fabric) * samps_per_clk
    σ = o["sigma"] * samps_per_clk * f_fabric # keep float!
    if o["style"] == "const":
        o["idata"] = None
    if o["style"] in ["flat_top", "stage"]: # l computed from σ
        l = int(σ * 5 / samps_per_clk / 2) * 2 * samps_per_clk
    if o["style"] in ["gaussian", "flat_top", "DRAG"]:
        x = np.arange(0, l)
        μ = l/2 - 0.5
        y = np.exp(-((x - μ) / σ) ** 2)
        o["idata"] = (y - y[0]) / (1 - y[0])
    if o["style"] == "stage":
        o["idata"] = np.zeros(int(σ * 4))
        for s in o["stage"]:
            o["idata"] = np.append(o["idata"], np.zeros(int(s[1] * f_fabric * samps_per_clk)) + s[0])
        o["idata"] = np.append(o["idata"], np.zeros(int(σ*4)))
        # pad o["idata"] to be integer multiple of samps_per_clk
        o["idata"] = np.append(o["idata"], np.zeros(samps_per_clk - len(o["idata"]) % samps_per_clk))
        o["idata"] = gaussian_filter(o["idata"], σ)
    if o["style"] == "DRAG":
        δ = -o["delta"] / (samps_per_clk * f_fabric)
        o["qdata"] = 0.5 * (x - μ) / (2 * σ ** 2) * o["idata"] / δ
    if o.get("idata") is not None:
        o["idata"] = np.array(o["idata"]) * (1 - o["ibias"]) + o["ibias"]
    if o.get("qdata") is not None:
        o["qdata"] = np.array(o["qdata"]) * (1 - o["qbias"]) + o["qbias"]

def parse(soccfg, cfg):
    """
    Creates the internal control object which controls the pulse sequence.
    c = {
        "p": {} # pulses
        "g": {} # Generator channels
        "r": {} # Readout channels
        "exec": [] # Execution list
    }
    """
    c = { "p": {}, "g": {}, "r": {}, "exec": [] }
    steps = cfg.pop("steps", []) # flatten steps
    for i in range(len(steps)):
        for key in steps[i]:
            if not f"{i}_{key}" in cfg:
                cfg[f"{i}_{key}"] = steps[i][key]
    for i in range(99999999): # find all execution steps
        if not f"{i}_type" in cfg:
            break
        o = { "type": cfg[f"{i}_type"] }
        o["t"] = cfg.get(f"{i}_t", 0)
        if o["type"] == "pulse":
            o["g"], o["p"], o["r"] = cfg.get(f"{i}_g"), cfg.get(f"{i}_p"), cfg.get(f"{i}_r", 0)
            c["g"][o["g"]] = { "p": o["p"], "r": None }
            c["p"][o["p"]] = { "g": o["g"], "r": None }
            o["threshold"] = cfg.get(f"{i}_threshold")
        if o["type"] == "trigger":
            o["rs"], o["p"] = cfg.get(f"{i}_rs"), cfg.get(f"{i}_p")
            if o["rs"] != None:
                for r in o["rs"]:
                    c["r"][r] = {}
        if o["type"] == "goto":
            o["rep"] = cfg.get(f"{i}_rep", 0)
            o["i"] = cfg.get(f"{i}_i", i + 1)
        c["exec"].append(o)
    for p in c["p"]: # find all used pulses
        o = c["p"][p]
        c["g"][o["g"]]["nqz"] = cfg.get(f"p{p}_nqz", 2)
        o["freq"] = cfg.get(f"p{p}_freq", 0)
        c["g"][o["g"]]["freq"] = o["freq"]
        c["g"][o["g"]]["mixer"] = c["g"][o["g"]].get("mixer", cfg.get(f"p{p}_mixer", None))
        o["mode"] = cfg.get(f"p{p}_mode", "oneshot")
        o["phrst"] = cfg.get(f"p{p}_phrst", 0)
        o["style"] = cfg.get(f"p{p}_style", "const")
        o["length"] = cfg.get(f"p{p}_length", 2)
        o["sigma"] = cfg.get(f"p{p}_sigma", o["length"] / 5)
        o["delta"] = cfg.get(f"p{p}_delta", -200)
        o["gain"] = cfg.get(f"p{p}_gain", 0)
        o["mask"] = cfg.get(f"p{p}_mask", list(range(len(o["freq"]))) if np.iterable(o["freq"]) else None)
        o["phase"] = cfg.get(f"p{p}_phase", list(np.zeros(len(o["freq"]))) if np.iterable(o["freq"]) else 0)
        o["stage"] = cfg.get(f"p{p}_stage", [])
        o["idata"] = np.array(cfg[f"p{p}_idata"]) if cfg.get(f"p{p}_idata") is not None else None
        o["qdata"] = np.array(cfg[f"p{p}_qdata"]) if cfg.get(f"p{p}_qdata") is not None else None
        o["ibias"] = cfg.get(f"p{p}_ibias", 0)
        o["qbias"] = cfg.get(f"p{p}_qbias", 0)
        if cfg.get(f"p{p}_power", None) is not None:
            o["gain"] = dB2gain(cfg[f"p{p}_power"])
            cfg[f"p{p}_gain"] = o["gain"] # write back computed gain
        if o["style"] == "flat_top": # not supporting mode
            o["mode"] = "oneshot"
        generate_waveform(o, soccfg)
    for r in range(10): # find all readout channels
        if f"r{r}_p" in cfg or f"r{r}_freq" in cfg:
            c["r"][r] = o = { "g": None }
            o["p"] = cfg.get(f"r{r}_p")
            o["freq"] = cfg.get(f"r{r}_freq")
            o["phase"] = cfg.get(f"r{r}_phase", 0)
            o["length"] = cfg.get(f"r{r}_length", 2)
            if o["p"] is not None: # match corresponding pulse/channels
                o["g"] = c["p"][o["p"]]["g"]
                c["g"][o["g"]]["r"] = c["p"][o["p"]]["r"] = r
                if o["freq"] is None:
                    o["freq"] = c["p"][o["p"]]["freq"]
    cfg["rep"] = cfg.get("rep", 0)
    return c

class Mercator(AveragerProgramV2):
    """General class for preparing and sending a pulse sequence."""
    def _initialize(self, cfg):
        c = self.c
        mux_gain_factor = { 1: 1, 2: 2, 3: 4, 4: 4, 5: 8, 6: 8, 7: 8, 8: 8 }
        for g, o in c["g"].items(): # Declare Generator Channels
            kwargs = {}
            if self.soccfg["gens"][g]["has_mixer"]:
                kwargs["ro_ch"] = o["r"]
                if np.iterable(o["freq"]): # mux
                    kwargs["mixer_freq"] = np.mean(o["freq"]) if o["mixer"] is None else o["mixer"]
                    kwargs["mux_freqs"] = o["freq"]
                    kwargs["mux_gains"] = np.array(c["p"][o["p"]]["gain"]) * mux_gain_factor[len(o["freq"])]
                    kwargs["mux_phases"] = c["p"][o["p"]]["phase"]
                else: # single tone
                    kwargs["mixer_freq"] = o["freq"] if o["mixer"] is None else o["mixer"]
            self.declare_gen(ch=g, nqz=o["nqz"], **kwargs)
        for r, o in c["r"].items(): # Declare Readout Channels
            kwargs = { "phase": o["phase"], "freq": o["freq"] }
            if o.get("g") is not None:
                kwargs["gen_ch"] = o["g"]
            if "tproc_ctrl" in self.soccfg["readouts"][r]:
                self.declare_readout(ch=r, length=o["length"])
                self.add_readoutconfig(ch=r, name=f"r{r}", **kwargs)
                self.send_readoutconfig(ch=r, name=f"r{r}", t=0)
            else: # mux readout
                self.declare_readout(ch=r, length=o["length"], **kwargs)
        for p, o in c["p"].items(): # Setup pulses
            kwargs = { "style": o["style"], "ro_ch": o["r"], "freq": o["freq"], "phase": o["phase"], "gain": o["gain"] }
            if o["mode"] == "last":
                kwargs["stdysel"] = "last"
                kwargs["mode"] = "oneshot"
            else:
                kwargs["mode"] = o["mode"]
            if not np.iterable(o["freq"]): # single tone
                kwargs["phrst"] = o["phrst"]
            if o["style"] == "const" and o["mask"] is not None: # mux mask
                kwargs["mask"] = o["mask"]
                for k in ["freq", "ro_ch", "mode", "phase", "gain"]:
                    kwargs.pop(k, None)
            if o["style"] != "const":
                kwargs["envelope"] = f"e{p}"
                maxv = self.soccfg.get_maxv(o["g"])
                idata = None if o["idata"] is None else maxv * np.array(o["idata"])
                qdata = None if o["qdata"] is None else maxv * np.array(o["qdata"])
                self.add_envelope(ch=o["g"], name=kwargs["envelope"], idata=idata, qdata=qdata)
            if o["style"] == "flat_top":
                kwargs.pop("mode", None) # no support
            if o["style"] in ["flat_top", "const"]:
                kwargs["length"] = o["length"]
            else:
                kwargs["style"] = "arb"
            self.add_pulse(ch=o["g"], name=f"p{p}",  **kwargs)
        if cfg["rep"] > 0:
            self.add_loop("rep", cfg["rep"])

    def _body(self, cfg):
        c = self.c
        goto_rep = {} # goto rep
        i = 0
        while i < len(c["exec"]):
            o = c["exec"][i]
            if o["type"] == "pulse":
                if o["threshold"] is not None: # conditional
                    self.read_and_jump(ro_ch=o["r"], component="I", threshold=self.us2cycles(o["threshold"] * self.c["r"][o["r"]]["length"], ro_ch=o["r"]), test="<", label=f"after_{i}")
                self.pulse(ch=o["g"], name=f"p{o['p']}", t=o["t"])
                if o["threshold"] is not None:
                    self.label(f"after_{i}")
            _wait_auto = lambda t: self.wait_auto(t, gens=True) # redefine to include gens
            time_functions = { "delay": self.delay, "delay_auto": self.delay_auto, "wait": self.wait, "wait_auto": _wait_auto }
            if o["type"] in time_functions:
                time_functions[o["type"]](o["t"])
            if o["type"] == "trigger":
                self.trigger(ros=(o["rs"] or self.c["r"].keys()), t=o["t"], pins=[0])
            if o["type"] == "goto":
                goto_rep[i] = goto_rep.get(i, o["rep"]) - 1
                i = o["i"] if goto_rep[i] >= 0 else i + 1
            else:
                i = i + 1

    def __init__(self, soccfg, cfg):
        self.c = parse(soccfg, cfg)
        super().__init__(soccfg, reps=cfg.get("hard_avg", 1), final_delay=1, cfg=cfg)

    def acquire(self, soc, progress=False, **kwargs):
        res = super().acquire(soc, progress=progress, rounds=self.cfg.get("soft_avg", 1), **kwargs)
        return np.moveaxis(res, -1, 0)

    def acquire_decimated(self, soc, progress=False, **kwargs):
        res = super().acquire_decimated(soc, progress=progress, rounds=self.cfg.get("soft_avg", 1), **kwargs)
        return np.moveaxis(res, -1, 0)

    def light(self):
        fig, ax = plt.subplots()
        c = self.c
        data = {} # plot data
        last = {} # last pulse in each generator
        for g in range(15, -1, -1):
            if g in c["g"]:
                data[g] = [[0, 0]]
                last[g] = { "mode": "oneshot", "end": 0 }
        delays = [0] # for plot, all t are absolute
        r_labeled = {} # for plot
        delay = 0 # current time reference
        wait = 0 # wait until time
        until = 0 # last pulse/readout end time
        def add_pulse(p, g, start, new_pulse=True): # return end time
            cycle = self.cycles2us(1, gen_ch=g)
            us = cycle / self.soccfg["gens"][g]["samps_per_clk"]
            if new_pulse and last[g]["mode"] == "periodic":
                while last[g]["end"] < start:
                    add_pulse(last[g]["p"], g, last[g]["end"] + cycle, new_pulse=False)
            start = max(start, last[g]["end"])
            if last[g]["mode"] == "last":
                data[g].append([start, last[g]["value"]])
            o = c["p"][p]
            last[g] = { "p": p, "mode": o["mode"] }
            gain = o["gain"][0] if np.iterable(o["gain"]) else o["gain"]
            if new_pulse:
                ax.annotate(f"p{p}", (start, gain + 0.05 if gain >= 0 else gain - 0.1))
            if o["idata"] is not None:
                l = o["length"] if o["style"] == "flat_top" else 0
                end = start + l + len(o["idata"]) * us
                h = len(o["idata"]) // 2
                data[g].extend(list(zip(np.arange(h)*us + start, gain * np.array(o["idata"])[:h])))
                data[g].extend(list(zip(np.arange(h)*us + start + l + h*us, gain * np.array(o["idata"])[h:])))
                if o["mode"] == "last":
                    last[g]["value"] = gain * o["idata"][-1]
            else:
                end = start + o["length"]
                if new_pulse:
                    data[g].append([start, 0])
                data[g].extend([[start, gain], [end, gain]])
                if o["mode"] == "last":
                    last[g]["value"] = gain
            if o["mode"] not in ["last", "periodic"]:
                data[g].append([end, 0])
            last[g]["end"] = end
            return end
        goto_rep = {}
        i = 0
        while i < len(c["exec"]):
            o = c["exec"][i]
            start = delay + o["t"]
            end = start
            if o["type"] == "pulse":
                start = max(start, wait, last[o["g"]]["end"])
                end = add_pulse(o["p"], o["g"], start)
                until = max(until, end)
            if o["type"] == "delay":
                delay = delay + o["t"]
                delays.append(delay)
            if o["type"] == "delay_auto":
                delay = max(until, delay) + o["t"]
                delays.append(delay)
            if o["type"] == "wait": # very confusing, but that's the behavior
                wait = delay + o["t"]
            if o["type"] == "wait_auto": # also confusing
                wait = until + o["t"]
            if o["type"] == "trigger":
                for r in (o["rs"] or c["r"]):
                    end = start + c["r"][r]["length"]
                    until = max(end, until)
                    ax.axvspan(xmin=start, xmax=end, color=('r' if r == 0 else 'b'), alpha=0.1, label=(None if r_labeled.get(r) else f"r{r}"))
                    r_labeled[r] = True
            if o["type"] == "goto":
                goto_rep[i] = goto_rep.get(i, o["rep"]) - 1
                i = o["i"] if goto_rep[i] >= 0 else i + 1
            else:
                i = i + 1
        final = max(until, delay)
        for g in data:
            cycle = self.cycles2us(1, gen_ch=g)
            if last[g]["mode"] == "periodic":
                while last[g]["end"] < final:
                    add_pulse(last[g]["p"], g, last[g]["end"] + cycle, new_pulse=False)
            elif last[g]["mode"] == "last":
                data[g].append([final, last[g]["value"]])
            else:
                data[g].append([final, 0])
            xy = np.transpose(data[g])
            ax.plot(xy[0], xy[1], label=f"g{g}", linewidth=3)
        ax.scatter(np.array(delays), np.zeros(len(delays)) - 1, label="delay", marker="^")
        ax.legend()
        ax.set_title("Mercator Light")
        ax.set_xlabel("Time (μs)")
        ax.set_ylabel("Gain")
        ax.set_ylim([-1.05, 1.2])
        ax.grid()
        return fig



import os

def get_version():
    with open(os.path.join(os.path.dirname(__file__), 'VERSION')) as version_file:
        return version_file.read().strip()

__version__ = get_version()

from .helper import *
from .mercator import Mercator
import quick.experiment as experiment
import quick.auto as auto



import numpy as np
import quick.helper as helper
from quick import __version__
from .mercator import Mercator
import os, yaml

# global var & config for quick.experiment
configs = helper.load_yaml(os.path.join(os.path.abspath(os.path.dirname(__file__)), "experiment.yml"))
var = configs["var"]

class BaseExperiment:
    var = {}
    var_label = {}
    _var = {}
    def __init__(self, data_path=None, title="", soccfg=None, soc=None, var=None, **kwargs):
        self.key = self.__class__.__name__ # get class name as experiment key
        self.data_path = data_path
        self.title = title
        template_var = dict(configs["var"])
        template_var.update(self.var)
        self.var = template_var
        self.var.update(var or {})
        template_var_label = dict(configs["var_label"])
        template_var_label.update(self.var_label)
        self.var_label = template_var_label
        self.sweep = {}
        self.config_update = {}
        for k, v in kwargs.items():
            if k in self.var:
                self.var[k] = v
                if np.iterable(v) and not isinstance(v, str):
                    self.sweep[k] = v
                    self.var[k] = v[0]
            else:
                self.config_update[k] = v
        self.eval_config(self.var)
        self.soccfg, self.soc = helper.getSoc() if soccfg is None else (soccfg, soc)
        self.m = Mercator(self.soccfg, self.config)
    def eval_config(self, v):
        configStr = configs.get(self.key, "")
        self.config = yaml.safe_load(helper.evalStr(configStr, v, self._var)) or {}
        self.config.update(self.config_update)
    def prepare(self, indep_params=[], dep_params=[], dB=False, population=False): # prepare saver
        self.data = []
        indep_params = list(indep_params) # avoid modify in place
        for k in self.sweep:
            label = self.var_label.get(k, (k, ""))
            indep_params.append(label)
        dep_params = list(dep_params)
        if len(dep_params) == 0:
            if population:
                dep_params.append(("Population", ""))
            dep_params.extend([("Amplitude", "dB") if dB else ("Amplitude", ""), ("Phase", "rad"), ("I", ""), ("Q", "")])
        if self.data_path is not None:
            self.s = helper.Saver(f"({self.key})" + self.title, self.data_path, indep_params, dep_params, { "quick_experiment": self.key, "quick_version": __version__, "config": self.config, "var": self.var })
    def add_data(self, data): # add & save data
        self.data.extend(data)
        if self.data_path is not None:
            self.s.write_data(data)
    def acquire_S21(self, indep_list, dB=False, population=False): # acquire & add data
        self.m = Mercator(self.soccfg, self.config)
        I, Q = self.m.acquire(self.soc)
        S21 = (np.array(I) + 1j * np.array(Q)).flatten()
        dep_list = []
        if population:
            I = S21.real
            dep_list.append(float(I[I > self.var.get("r_threshold", 0)].size) / I.size)
        S21 = np.mean(S21)
        dep_list.extend([(20 * np.log10(np.abs(S21) / self.config["p0_gain"]) if dB else np.abs(S21)), np.angle(S21), S21.real, S21.imag])
        self.add_data([[*indep_list, *dep_list]])
    def run(self, silent=False, dB=False, population=False):
        if not silent:
            print(f"quick.experiment({self.key}) Starting")
        self.prepare(dB=dB, population=population)
        for v in helper.Sweep(self.var, self.sweep, progressBar=(not silent)):
            self.eval_config(v)
            indep = []
            for k in self.sweep:
                indep.append(v[k])
            self.var = v
            self.acquire_S21(indep, dB=dB, population=population)
        return self.conclude(silent)
    def conclude(self, silent=False): # last step of run
        if not silent:
            print(f"quick.experiment({self.key}) Completed. Data saved in {self.data_path}" if self.data_path is not None else f"quick.experiment({self.key}) Completed.")
        self.data = np.array(self.data)
        if self.data_path is not None:
            self.s.write_yml() # update the completed time
        return self
    def light(self):
        self.m.light()
        return self

# All experiments are following
class LoopBack(BaseExperiment):
    def run(self, silent=False, dB=False):
        if not silent:
            print(f"quick.experiment({self.key}) Starting")
        self.prepare(indep_params=[("Time", "us")], dB=dB)
        self.eval_config(self.var)
        self.m = Mercator(self.soccfg, self.config)
        I, Q = self.m.acquire_decimated(self.soc, progress=not silent)
        t = self.m.get_time_axis(ro_index=0)
        S21 = I[0] + 1j * Q[0]
        self.add_data(np.transpose([t, (20 * np.log10(np.abs(S21) / self.config["p0_gain"]) if dB else np.abs(S21)), np.angle(S21), S21.real, S21.imag]))
        return self.conclude(silent)

class ResonatorSpectroscopy(BaseExperiment):
    def run(self, silent=False, dB=True):
        return super().run(silent, dB)

class QubitSpectroscopy(BaseExperiment):
    pass # The BaseExperiment class is sufficiently powerful to do everything here.

class Rabi(BaseExperiment):
    def __init__(self, **kwargs):
        self.var = { "cycle": 0 } # add one var
        self.var_label = { "cycle": ("Extra Pi Pulse", "") }
        super().__init__(**kwargs)
    def run(self, silent=False, population=True):
        return super().run(silent=silent, population=population)

class IQScatter(BaseExperiment):
    def __init__(self, **kwargs):
        self.var = { "rr_length": None } # add one var
        self.var_label = { "rr_length": ("Readout Window Time", "us") }
        super().__init__(**kwargs)
    def run(self, silent=False):
        if not silent:
            print(f"quick.experiment({self.key}) Starting")
        self.prepare(dep_params=[("I 0", ""), ("Q 0", ""), ("I 1", ""), ("Q 1", "")])
        for v in helper.Sweep(self.var, self.sweep, progressBar=(not silent)):
            self.eval_config(v)
            self.var = v
            self.config["0_type"] = "pulse" # send pi pulse
            self.config["1_t"] = 0
            self.m = Mercator(self.soccfg, self.config)
            I1, Q1 = self.m.acquire(self.soc)
            I1, Q1 = np.array(I1[0][0]).flatten(), np.array(Q1[0][0]).flatten()
            self.config["0_type"] = "delay_auto" # omit pi pulse
            self.config["1_t"] = self.var["q_length"]
            self.m = Mercator(self.soccfg, self.config)
            I0, Q0 = self.m.acquire(self.soc)
            I0, Q0 = np.array(I0[0][0]).flatten(), np.array(Q0[0][0]).flatten()
            indep = []
            for k in self.sweep:
                indep.append(np.zeros(len(I0)) + v[k])
            self.add_data(np.transpose([*indep, I0, Q0, I1, Q1]))
        return self.conclude(silent)

class DispersiveSpectroscopy(BaseExperiment):
    def run(self, silent=False, dB=True):
        if not silent:
            print(f"quick.experiment({self.key}) Starting")
        self.prepare(dep_params=[("Amplitude 0", "dB" if dB else ""), ("Phase 0", "rad"), ("I 0", ""), ("Q 0", ""), ("Amplitude 1", "dB" if dB else ""), ("Phase 1", "rad"), ("I 1", ""), ("Q 1", "")])
        for v in helper.Sweep(self.var, self.sweep, progressBar=(not silent)):
            self.eval_config(v)
            indep = []
            for k in self.sweep:
                indep.append(v[k])
            self.var = v
            c = self.config
            c["0_type"] = "pulse" # send pi pulse
            c["1_t"] = 0
            self.m = Mercator(self.soccfg, c)
            I1, Q1 = self.m.acquire(self.soc)
            S1 = I1[0][0] + 1j * Q1[0][0]
            c["0_type"] = "delay_auto" # omit pi pulse
            c["1_t"] = self.var["q_length"]
            self.m = Mercator(self.soccfg, c)
            I0, Q0 = self.m.acquire(self.soc)
            S0 = I0[0][0] + 1j * Q0[0][0]
            self.add_data([[*indep, 20 * np.log10(np.abs(S0) / c["p0_gain"]) if dB else np.abs(S0), np.angle(S0), np.real(S0), np.imag(S0), 20 * np.log10(np.abs(S1) / c["p0_gain"]) if dB else np.abs(S1), np.angle(S1), np.real(S1), np.imag(S1) ]])
        return self.conclude(silent)

class T1(BaseExperiment):
    def __init__(self, **kwargs):
        self.var = { "time": 0 } # add one var
        self.var_label = { "time": ("Delay Time", "us") }
        super().__init__(**kwargs)
    def run(self, silent=False, population=True):
        return super().run(silent=silent, population=population)

class T2Ramsey(BaseExperiment):
    def __init__(self, **kwargs):
        self.var = { "time": 0, "fringe_freq": 0 } # add var
        self.var_label = {
            "time": ("Delay Time", "us"),
            "fringe_freq": ("Fringe Frequency", "MHz")
        }
        super().__init__(**kwargs)
    def run(self, silent=False, population=True):
        return super().run(silent=silent, population=population)

class T2Echo(BaseExperiment):
    def __init__(self, **kwargs):
        self.var = { "time": 0, "cycle": 0, "fringe_freq": 0 } # add var
        self.var_label = {
            "time": ("Delay Time", "us"),
            "cycle": ("Extra Cycles", ""),
            "fringe_freq": ("Fringe Frequency", "MHz")
        }
        super().__init__(**kwargs)
    def run(self, silent=False, population=True):
        return super().run(silent=silent, population=population)

class QND(BaseExperiment):
    def __init__(self, **kwargs):
        def get_steps(seq, v={}):
            res = ""
            for s in seq:
                if s > 0:
                    res += f"- type: pulse\n  p: 1\n  g: {v['q']}\n- type: delay_auto\n"
                else:
                    res += f"- type: delay_auto\n  t: {v['q_length']}\n"
                res += f"- type: pulse\n  p: 0\n  g: {v['r']}\n"
                res += f"- type: trigger\n  t: {v['r_offset']}\n"
                res += f"- type: delay_auto\n  t: {v['r_reset']}\n"
            return res
        self.var = { "cycle": 10, "random": 100 } # add var
        self.var_label = {
            "cycle": ("Cycles", ""),
            "random": ("Random", "")
        }
        self._var = { "get_steps": get_steps, "seq": [1] }
        super().__init__(**kwargs)
    def run(self, silent=False):
        if not silent:
            print(f"quick.experiment({self.key}) Starting")
        self.prepare(indep_params=[("Cycle Index", "")], dep_params=[("Correlation", "")])
        for v in helper.Sweep(self.var, self.sweep, progressBar=(not silent)):
            indep = []
            for k in self.sweep:
                indep.append(v[k] + np.zeros(v["cycle"]))
            self.var = v
            Cs = []
            for _ in range(v["random"]):
                self._var["seq"] = np.random.randint(2, size=v["cycle"])
                self.eval_config(v)
                self.m = Mercator(self.soccfg, self.config)
                I_list, _ = self.m.acquire(self.soc)
                R = (I_list[0] > v.get("r_threshold", 0)).astype(int) # {0, 1} readout
                R_shift = np.concatenate(([np.zeros(len(R[0]))], R[:-1])).astype(int)
                O = np.bitwise_xor(R, R_shift).astype(int) # {0, 1} 1 if state changes
                O = O[:, O[0, :] < 1] # post-select state-prep in g
                O = np.mean(O, axis=1)[1:] # average across rep, ignore first readout
                C = 1 - 2 * np.abs(self._var["seq"] - O) # correlation with pi pulse
                Cs.append(C)
            C = np.mean(Cs, axis=0)
            self.add_data(np.transpose([np.arange(0, v["cycle"]), *indep, C ]))
        return self.conclude(silent)



from setuptools import setup, find_packages
import pathlib

here = pathlib.Path(__file__).parent.resolve()
def get_version(rel_path):
    return (here / rel_path).read_text().strip()

setup(
    name="clelandlab-quick",
    version=get_version("quick/VERSION"),
    description="QuICK is a universal wrap of QICK.",
    long_description=(here / "README.md").read_text(encoding="utf-8"),
    long_description_content_type="text/markdown",
    url="https://github.com/clelandlab/quick",
    author="Cleland Lab",
    author_email="clelandlab@proton.me",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Topic :: Software Development :: Build Tools",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Programming Language :: Python :: 3 :: Only",
    ],
    keywords="QICK,quantum,experiment,measurement,qubit,control,readout,fpga",
    packages=find_packages(where="."),
    python_requires=">=3.8, <4",
    install_requires=[
        "qick==0.2.366",
        "numpy",
        "scipy",
        "pyyaml",
        "pyro4",
        "matplotlib",
        "ipython"
    ],
    package_data={
        "quick": ["VERSION", "experiment.yml"],
    },
    include_package_data=True,
    project_urls={
        "Source": "https://github.com/clelandlab/quick",
        "Documentation": "https://clelandlab-quick.readthedocs.io/en/latest/",
        "Tracker": "https://github.com/clelandlab/quick/issues",
    },
)



import numpy as np
import matplotlib.pyplot as plt
from multiprocessing import Process, Queue
from scipy.optimize import minimize, curve_fit, leastsq
from scipy import interpolate, stats
import yaml, json, os, re, ast
from tqdm.notebook import tqdm
from datetime import datetime
import Pyro4
from qick import QickConfig

_soccfg, _soc = None, None
π = np.pi

yaml.add_representer(np.ndarray, lambda dumper, array: dumper.represent_sequence('tag:yaml.org,2002:seq', array.tolist(), flow_style=True))
yaml.add_multi_representer(np.generic, lambda dumper, data: dumper.represent_data(data.item()))

def connect(ip, port=8888, proxy_name="qick"):
    global _soc, _soccfg
    Pyro4.config.SERIALIZER = "pickle"
    Pyro4.config.PICKLE_PROTOCOL_VERSION = 4
    ns = Pyro4.locateNS(host=ip, port=port)
    _soc = Pyro4.Proxy(ns.lookup(proxy_name))
    _soc.get_cfg()
    _soccfg = QickConfig(_soc.get_cfg())
    return _soccfg, _soc

def getSoc(): # Return cached connection
    return _soccfg, _soc

def print_yaml(data): # Print in yaml format
    print(yaml.dump(data))

def load_yaml(path):
    file = open(path, "r")
    content = yaml.safe_load(file)
    file.close()
    return content

def save_yaml(path, data):
    file = open(path, "w")
    content = yaml.dump(data)
    file.write(content)
    return content

def load_data(*paths):
    data_list = []
    for p in paths:
        d = np.genfromtxt(p, delimiter=",")
        data_list.append(d)
    return np.concatenate(data_list)

def feistel_network(i, N, seed): # generate pseudo-random permutation
    half_bits = (N - 1).bit_length() // 2 + (N - 1).bit_length() % 2
    limit = 1 << half_bits
    mask = limit - 1
    def round_function(val, round_idx): # a variation of a 32-bit MurmurHash-style mixer.
        x = (val ^ seed ^ round_idx) * 0x45d9f3b
        x = ((x >> 16) ^ x) * 0x45d9f3b
        x = (x >> 16) ^ x
        return x & mask
    def permute(val):
        left = val >> half_bits
        right = val & mask
        for r in range(5):
            left, right = right, left ^ round_function(right, r)
        return (left << half_bits) | right
    curr = permute(i % N)
    while curr >= N:
        curr = permute(curr)
    return curr

def nested_get(d, key): # key is a list
    for k in key:
        d = d[k]
    return d

class Sweep:
    def __init__(self, config, sweepConfig, random=False, progressBar=True):
        self.config = dict(config) # Copy config, avoid changing the original
        self.sweep = [] # [{ "key", "list", "len" }]
        self.progressBar = progressBar
        self.random = random
        self.total = 1
        self.index = 0 # iteration index, increase in order
        self.seed = np.random.randint(2147483647)
        def parse_sweep(key, d): # key is a list
            for k, v in d.items():
                if isinstance(v, dict):
                    parse_sweep(key + [k], v)
                    continue
                if np.iterable(v) and not isinstance(v, str):
                    self.sweep.append({ "key": key + [k], "list": list(v), "len": len(v) })
                    self.total *= len(v)
                    continue
                nested_get(self.config, key)[k] = v
        try:
            parse_sweep([], sweepConfig)
            self.sweep.reverse()
            if self.total == 0:
                raise("Empty Sweep")
        except:
            raise("Invalid Sweep Config")
    def __iter__(self): # initialze all iterator
        if self.progressBar:
            self.progress = tqdm(total=self.total, mininterval=0, miniters=1, desc='quick.Sweep')
        self.index = 0
        return self
    def __next__(self):
        if self.index == self.total:
            if self.progressBar:
                self.progress.close()
            raise StopIteration
        if self.progressBar:
            self.progress.update()
        i = feistel_network(self.index, self.total, self.seed) if self.random else self.index
        for s in self.sweep:
            nested_get(self.config, s["key"][:-1])[s["key"][-1]] = s["list"][i % s["len"]]
            i = i // s["len"]
        self.index += 1
        return self.config

class Saver:
    def __init__(self, title, path, indep_params=[], dep_params=[], params={}):
        self.indep_params = []
        self.dep_params = []
        self.params = params
        self.title = title
        self.path = path
        self.file_name = ''
        self.has_data = False
        os.path.exists(path)
        try: # check variables
            for value in indep_params:
                self.indep_params.append([value[0], value[1]])
            for value in dep_params:
                self.dep_params.append([value[0], value[1]])
        except KeyError:
            print("Variables incorrectly formatted. Variable: ('label', 'unit')")
        # Detect max existing file number
        existing_files = [f for f in os.listdir(self.path) if re.search(r"^\d\d\d\d\d \- .*\.csv$", f)]
        if existing_files:
            existing_numbers = [int(f.split(' - ')[0]) for f in existing_files]
            next_number = max(existing_numbers) + 1
        else:
            next_number = 0
        # Creating file name + path
        self.file_name = self.path + f"/{next_number:05d} - {self.title}"
        self.created_time = datetime.now().strftime("%Y-%m-%d %H:%M:%S")
        self.write_yml()
        np.savetxt(self.file_name + '.csv', [], delimiter=',')
    def write_yml(self):
        meta = {
            "created": self.created_time,
            "completed": datetime.now().strftime("%Y-%m-%d %H:%M:%S") if self.has_data else "N/A",
            "title": self.title,
            "independent": self.indep_params,
            "dependent": self.dep_params,
            "parameters": self.params
        }
        save_yaml(self.file_name + ".yml", meta)
    def write_data(self, data):
        self.has_data = True
        with open(self.file_name + '.csv', 'a+') as f:
            np.savetxt(f, data, fmt="%.9e", delimiter=',')
        # self.write_yml() # do not call it every time for efficiency

def dB2gain(dB, ref_gain=1, ref_dB=None):
    return ref_gain * 10 ** (dB / 20) if ref_dB is None else 10 ** ((dB + ref_dB) / 20)

def gain2dB(gain, ref_gain=1, ref_dB=None):
    return 20 * np.log10(gain / ref_gain) if ref_dB is None else 20 * np.log10(gain) - ref_dB

def evalStr(s, var, _var=None):
    return eval(f"f'''{s}'''", _var, var)

def safe_wrap(retry=3, timeout=300):
    def decorator(f):
        def _run(q, *a, **kw):
            try:
                q.put(('R', f(*a, **kw)))
            except Exception as e:
                q.put(('E', e))
        def wrap(*a, **kw):
            last_error = TimeoutError("All retries and timeouts exceeded")
            for _ in range(retry + 1):
                q = Queue()
                p = Process(target=_run, args=(q,) + a, kwargs=kw)
                p.start()
                p.join(timeout)
                if p.is_alive():
                    p.terminate()
                    last_error = TimeoutError("Function timeout")
                elif not q.empty():
                    s, res = q.get()
                    if s == 'R': return res
                    last_error = res
            raise last_error
        return wrap
    return decorator

def symmetryCenter(x, y, it=3):
    L = len(y)
    R = int(L / 2)
    Δ = int(R / 2)
    freq = np.fft.rfftfreq(L*2)
    y = (y - np.mean(y)) / np.std(y)
    fft = np.fft.rfft(np.append(y, np.zeros(L)))
    res, d = 0, 0
    for _ in range(it):
        fft *= np.exp(2j*π*d*freq)
        co = np.fft.irfft(fft**2)[R-1:L+R-1]
        i = np.argmax(co[R-Δ:R+Δ]) + R-Δ
        p = np.polynomial.polynomial.polyfit(np.arange(i-2, i+3), co[i-2:i+3], 2)
        d = -p[1]/4/p[2] - R/2
        res += d
    i = res + R - 0.5
    _i = int(i)
    return x[_i] + (x[_i+1] - x[_i]) * (i - _i)

def estimateOmega(x, y):
    y = (y - np.mean(y)) / np.std(y)
    rfft = np.fft.rfft(y)
    freq = np.fft.rfftfreq(len(y))
    return len(y) * freq[np.argmax(np.abs(rfft))] * 2 * π / (x[len(x) - 1] - x[0])

def iq_scatter(S0s, S1s, c0=None, c1=None, plot=True):
    if c0 is None:
        c0 = np.median(S0s.real) + 1j * np.median(S0s.imag)
    if c1 is None:
        c1 = np.median(S1s.real) + 1j * np.median(S1s.imag)
    def negative_visibility(_p):
        p = _p[0] + 1j * _p[1]
        angle = np.angle(c0 - p)
        S0 = (S0s - (p + c0) / 2.0) * np.exp(-1j * angle)
        S1 = (S1s - (p + c0) / 2.0) * np.exp(-1j * angle)
        prob0 = float(S0.real[S0.real < 0].size) / S0.size
        prob1 = float(S1.real[S1.real < 0].size) / S1.size
        return prob0 - prob1
    res = minimize(negative_visibility, [c1.real, c1.imag], method='Nelder-Mead') # optimize c1
    visibility = -1 * negative_visibility(res.x)
    c1 = res.x[0] + 1j * res.x[1]
    o = (c0 + c1) / 2.0 # origin
    g = o - c0 # project on opposite of g
    S0p = np.real((S0s - o) * np.conj(g)) / np.abs(g)
    S1p = np.real((S1s - o) * np.conj(g)) / np.abs(g)
    S0p_right, S1p_right = S0p[S0p < 0], S1p[S1p > 0]
    S0s_right, S0s_wrong = S0s[S0p < 0], S0s[S0p >= 0]
    S1s_right, S1s_wrong = S1s[S1p > 0], S1s[S1p <= 0]
    Fg, Fe = len(S0p_right) / len(S0p), len(S1p_right) / len(S1p)
    phase, threshold = float(-np.angle(c1 - c0, deg=True)), float(np.real((c1 + c0) / 2 * np.exp(-1j *  np.angle(c1 - c0))))
    if not plot:
        return phase, threshold, visibility, Fg, Fe, c0, c1, None
    fig, axes = plt.subplots(1, 2, figsize=(12, 5.44)) # plot the histogram
    xp = np.linspace(np.min(S0p), np.max(S1p), 2001) # axis of projection
    g_hist, _ = np.histogram(S0p, bins=100, density=True, range=[xp[0], xp[-1]])
    e_hist, _ = np.histogram(S1p, bins=100, density=True, range=[xp[0], xp[-1]])
    g_cdf = np.cumsum(g_hist) * (xp[-1] - xp[0]) / 100
    e_cdf = np.cumsum(e_hist) * (xp[-1] - xp[0]) / 100
    axes[1].hist(S0p, color='blue', alpha=0.5, bins=100, density=True, range=[xp[0], xp[-1]], label=r'$|g\rangle$')
    axes[1].hist(S1p, color='red', alpha=0.5, bins=100, density=True, range=[xp[0], xp[-1]], label=r'$|e\rangle$')
    g_mu, g_std = stats.norm.fit(S0p_right)
    axes[1].plot(xp, stats.norm.pdf(xp, g_mu, g_std), color='blue', linestyle='dotted')
    e_mu, e_std = stats.norm.fit(S1p_right)
    axes[1].plot(xp, stats.norm.pdf(xp, e_mu, e_std), color='red', linestyle='dotted')
    axr = axes[1].twinx()
    axr.set_ylim([0, 1])
    axr.set_ylabel("Cumulative Probability")
    axr.plot(np.linspace(xp[0], xp[-1], 100), g_cdf, color='blue', linewidth=2)
    axr.plot(np.linspace(xp[0], xp[-1], 100), e_cdf, color='red', linewidth=2)
    axr.plot(np.linspace(xp[0], xp[-1], 100), g_cdf - e_cdf, color='green', linewidth=2)
    axes[1].set_xlabel('Projected IQ Amplitude (a.u.)')
    axes[1].set_ylabel('Probability Density')
    axes[1].legend(shadow=False, loc=1, frameon=True)
    axes[1].grid()
    axes[1].text(0.05, 0.95, 'Visibility=%.3f' % visibility, ha='left', transform=axes[1].transAxes)
    axes[1].text(0.05, 0.9, '$F_g$=%.3f' % Fg, ha='left', transform=axes[1].transAxes)
    axes[1].text(0.05, 0.85, '$F_e$=%.3f' % Fe, ha='left', transform=axes[1].transAxes)
    # scatter plot
    axes[0].scatter(S0s_right.real, S0s_right.imag, s=10, marker='o', zorder=0, alpha=0.5, c='blue', linewidths=0.0)
    axes[0].scatter(S1s_right.real, S1s_right.imag, s=10, marker='o', zorder=0, alpha=0.5, c='red', linewidths=0.0)
    axes[0].scatter(S0s_wrong.real, S0s_wrong.imag, s=10, marker='o', zorder=0, alpha=0.5, c='blue', linewidths=0.0)
    axes[0].scatter(S1s_wrong.real, S1s_wrong.imag, s=10, marker='o', zorder=0, alpha=0.5, c='red', linewidths=0.0)
    axes[0].grid()
    axes[0].set_xlabel('I (a.u.)')
    axes[0].set_ylabel('Q (a.u.)')
    _r = max(max(np.max(np.real(S0s - o)), np.max(np.real(S1s - o))), max(np.max(np.imag(S0s - o)), np.max(np.imag(S1s - o))))
    axes[0].set_xlim(-1.1 * _r + o.real, 1.1 * _r + o.real)
    axes[0].set_ylim(-1.1 * _r + o.imag, 1.1 * _r + o.imag)
    # plot the cut
    _x = np.linspace(-1.1 * _r + o.real, 1.1 * _r + o.real, 101)
    axes[0].plot(_x, -(c0.real - c1.real) / (c0.imag - c1.imag) * (_x - (c0.real + c1.real) / 2) + (c0.imag + c1.imag) / 2, 'k--')
    # add the circle to the IQ scatter plot
    theta = np.linspace(-π, π, 1001)
    I_g_fit_list = np.median(S0s_right.real) + np.cos(theta) * 3 * g_std
    Q_g_fit_list = np.median(S0s_right.imag) + np.sin(theta) * 3 * g_std
    I_e_fit_list = np.median(S1s_right.real) + np.cos(theta) * 3 * e_std
    Q_e_fit_list = np.median(S1s_right.imag) + np.sin(theta) * 3 * e_std
    axes[0].plot(I_g_fit_list, Q_g_fit_list, color='darkblue', linestyle='--', label=r'$|g\rangle$, $3\sigma$')
    axes[0].plot(I_e_fit_list, Q_e_fit_list, color='darkred', linestyle='--', label=r'$|e\rangle$, $3\sigma$')
    axes[0].legend(shadow=False, loc=1, frameon=True)
    return phase, threshold, visibility, Fg, Fe, c0, c1, fig

# T1 fit and plot
def fitT1(T, S, plot=True):
    def m(x, *p):
        return p[0] * np.exp(-x / p[1]) + p[2]
    def jac(x, *p):
        d0 = np.exp(-x / p[1])
        d1 = p[0] * x / p[1] ** 2 * np.exp(-x / p[1])
        d2 = np.ones(len(x))
        return np.transpose([d0, d1, d2])
    p0 = [1.0, 1.0, 1.0]
    p, pcov = curve_fit(m, T, S, p0=p0, jac=jac)
    perr = np.sqrt(np.diag(pcov))  # Standard deviation of parameters
    residuals = S - m(T, *p)
    r2 = 1 - np.sum(residuals**2) / np.sum((S - np.mean(S))**2)
    if not plot:
        return p, perr, r2, None
    dof = len(T) - len(p)
    fig, ax = plt.subplots()
    ax.scatter(T, S, color='black', s=10, label='Original Data')
    ax.plot(T, m(T, *p), color='red', label='Fit')
    annotation_text = (
        r"$S = S_0e^{-\tau/T_1} + C$" + "\n" +
        "Fitting Result:\n" +
        r"$S_0 = ({:.2f} \pm {:.2f})$".format(p[0], perr[0]) + "\n" +
        r"$T_1 = ({:.2f} \pm {:.2f})$ us".format(p[1], perr[1]) + "\n" +
        r"$C = ({:.2f} \pm {:.2f})$".format(p[2], perr[2]) + "\n" +
        r"$n = {}$ (DOF)".format(dof) + "\n" +
        r"$R^2 = {:.3f}$%".format(r2*100)
    )
    ax.set_xlabel(r"Pulse Delay $\tau$ (us)")
    ax.set_ylabel("Population")
    ax.set_title("T1 Measurement")
    ax.grid()
    ax.legend()
    mid_index = len(T) // 2
    ax.annotate(annotation_text, (T[mid_index], S[mid_index]), fontsize=8, xycoords='data', textcoords='offset points', xytext=(20, 20))
    return p, perr, r2, fig

# T2 fit and plot
def fitT2(T, S, omega=2*π, T2=20.0, plot=True):
    def m(x, *p):
        return p[0] * np.exp(-x / p[1]) * np.cos(p[2] * x) + p[3]
    def jac(x, *p):
        d0 = np.exp(-x / p[1]) * np.cos(p[2] * x)
        d1 = p[0] * x / p[1] ** 2 * np.exp(-x / p[1]) * np.cos(p[2] * x)
        d2 = - x * p[0] * np.exp(-x / p[1]) * np.sin(p[2] * x)
        d3 = np.ones(len(x))
        return np.transpose([d0, d1, d2, d3])
    def me(x, *p):
        return p[0] * np.exp(-x / p[1]) + p[3]
    p0 = [0.4, T2, omega, 0.5]
    p, pcov = curve_fit(m, T, S, p0=p0, jac=jac)
    perr = np.sqrt(np.diag(pcov))  # Standard deviation of parameters
    residuals = S - m(T, *p)
    r2 = 1 - np.sum(residuals**2) / np.sum((S - np.mean(S))**2)
    if not plot:
        return p, perr, r2, None
    dof = len(T) - len(p)
    fig, ax = plt.subplots()
    ax.scatter(T, S, color='black', s=10, label='Original Data')
    ax.plot(T, m(T, *p), color='red', label='Fit')
    ax.plot(T, me(T, *p), color='blue', label='Fit Exponential')
    annotation_text = (
        r"$S = S_0e^{-\tau/T_2}\cos(\omega\tau) + C$" + "\n" +
        "Fitting Result:\n" +
        r"$S_0 = ({:.2f} \pm {:.2f})$, $C = ({:.2f} \pm {:.2f})$".format(p[0], perr[0], p[3], perr[3]) + "\n" +
        r"$T_2 = ({:.2f} \pm {:.2f})$ us".format(p[1], perr[1]) + "\n" +
        r"$\omega = ({:.2f} \pm {:.2f})$ rad/us".format(p[2], perr[2]) + "\n" +
        r"$n = {}$ (DOF), $R^2 = {:.3f}$%".format(dof, r2 * 100)
    )
    ax.set_xlabel(r"Pulse Delay $\tau$ (us)")
    ax.set_ylabel("Population")
    ax.set_title("T2 Measurement")
    ax.grid()
    ax.legend(loc="lower right")
    mid_index = len(T) // 2
    ax.annotate(annotation_text, (T[mid_index], me(T[mid_index], *p)), fontsize=8, xycoords='data', textcoords='offset points', xytext=(10, 10))
    return p, perr, r2, fig

def fitResonator(F, S, fit="circle", p0=[None, None, None, None], plot=True, normalize_background=True, normalize_phase=True):
    def dB(a):
        return 20. * np.log10(np.abs(a))
    def S21_th(f, *p):
        Qi, Qc, fr, phi = p
        return 1 / (1 + Qi / Qc * np.exp(1j * phi) / (1 + 2j * Qi * (f - fr) / fr))
    def normalize(F, S):
        n = 4
        s = F.argsort()
        F, S = F[s], S[s]
        logmag = dB(S)
        if normalize_background:
            magoffset = np.mean(np.concatenate((logmag[:n], logmag[-n:])))
            logmag = logmag - magoffset
            fit_logmag_background = np.polyfit(np.concatenate((F[:n], F[-n:])), np.concatenate((logmag[:n], logmag[-n:])), 1)
            logmag_background_line = np.poly1d(fit_logmag_background)
            logmag = logmag - logmag_background_line(F)
        phase = np.unwrap(np.angle(S))
        if normalize_phase:
            fitphase = np.polyfit(np.concatenate((F[:n], F[-n:])), np.concatenate((phase[:n], phase[-n:])), 1)
            phaseline = np.poly1d(fitphase)
            phase = phase - phaseline(F)
        S = 10**(logmag/20.)*np.exp(1j*phase)
        return F, S
    F, S = normalize(F, S)
    S_inv, S_dB = 1 / S, dB(S)
    _p0 = [100000, 10000, F[np.argmin(S_dB)], 0.0]
    p0 = list(p0)
    for i in range(len(p0)):
        if p0[i] is None:
            p0[i] = _p0[i]
    bounds = ([-np.inf, -np.inf, 0, -2*π], [np.inf, np.inf, np.inf, 2*π])
    def p_b2i(p):
        _p = []
        for i in range(len(p)):
            if bounds[0][i] == -np.inf:
                _p.append(p[i])
                continue
            if bounds[1][i] == np.inf:
                _p.append(np.sqrt((p[i] - bounds[0][i] + 1) ** 2 - 1))
                continue
            _p.append(np.arcsin(2 * (p[i] - bounds[0][i]) / (bounds[1][i] - bounds[0][i]) - 1))
        return _p
    def p_i2b(_p):
        p = []
        for i in range(len(_p)):
            if bounds[0][i] == -np.inf:
                p.append(_p[i])
                continue
            if bounds[1][i] == np.inf:
                p.append(bounds[0][i] - 1 + np.sqrt(_p[i] ** 2 + 1))
                continue
            p.append(bounds[0][i] + (np.sin(_p[i]) + 1) * (bounds[1][i] - bounds[0][i]) / 2)
        return p
    def err_i2b(_p, _perr):
        perr = []
        for i in range(len(_p)):
            if bounds[1][i] == np.inf:
                perr.append(np.abs(_p[i] / np.sqrt(_p[i]**2 + 1)) * _perr[i])
            else:
                perr.append(np.abs(np.cos(_p[i]) * (bounds[1][i] - bounds[0][i]) / 2) * _perr[i])
        return perr
    def S21_circle_fit_err(_p):
        y = 1 / S21_th(F, *p_i2b(_p))
        return np.concatenate((np.real(S_inv - y), np.imag(S_inv - y)))
    def S21_amp_fit_err(_p):
        return dB(S21_th(F, *p_i2b(_p))) - S_dB
    def S21_arg_fit_err(_p):
        return np.angle(S / S21_th(F, *p_i2b(_p)))
    fun = { "circle": S21_circle_fit_err, "amp": S21_amp_fit_err, "arg": S21_arg_fit_err }
    result, cov, _, _, _ = leastsq(fun[fit], p_b2i(p0), maxfev=50000, xtol=1.e-7, ftol=1.e-7, col_deriv=False, gtol=1.e-7, full_output=True)
    p = p_i2b(result)
    perr = err_i2b(result, np.sqrt(np.diag(cov*np.var(fun[fit](p)))))
    S21_fit = S21_th(F, *p)
    residuals = 1/S - 1/S21_fit
    residuals = np.concatenate((np.real(residuals), np.imag(residuals)))
    flat_S_inv = np.concatenate((np.real(S_inv), np.imag(S_inv)))
    r2 = 1 - np.sum(residuals**2) / np.sum((flat_S_inv - np.mean(flat_S_inv))**2)
    p[0], p[1] = abs(p[0]), abs(p[1])
    if not plot:
        return p, perr, r2, None
    dof = len(residuals) - len(p)
    fig = plt.figure(figsize=(12, 6), tight_layout=True)
    axes = [fig.add_subplot(121), fig.add_subplot(222), fig.add_subplot(224)]
    axes[0].scatter((1 / S).real, (1 / S).imag, label="raw", alpha=0.5, linewidths=0.0)
    axes[0].plot((1 / S21_fit).real, (1 / S21_fit).imag, "r-", label="fit")
    axes[0].set_xlabel(r"Re $\widetilde{S}_{21}^{-1}$")
    axes[0].set_ylabel(r"Im $\widetilde{S}_{21}^{-1}$")
    axes[0].grid()
    axes[1].scatter(F, S_dB, alpha=0.5, linewidths=0.0)
    axes[1].plot(F, dB(S21_fit), "r-")
    axes[1].set_ylabel(r"$|S_{21}|$(dB)")
    axes[1].text(0.025, 0.21, r"$Q_i = %.3e$" % p[0] + "\n" + r"        $\pm %.1e$" % perr[0], ha="left", va="bottom", transform=axes[1].transAxes)
    axes[1].text(0.025, 0.05, r"$Q_c = %.3e$" % p[1] + "\n" + r"        $\pm %.1e$" % perr[1], ha="left", va="bottom", transform=axes[1].transAxes)
    axes[1].text(0.7, 0.05, r"$f = %.7e$" % p[2], ha="left", va="bottom", transform=axes[1].transAxes)
    axes[1].grid()
    mean_phase_data, mean_phase_fit = np.mean(np.unwrap(np.angle(S))), np.mean(np.unwrap(np.angle(S21_fit)))
    match = int((mean_phase_fit - mean_phase_data) / (2 * π) + 0.5)
    axes[2].sharex(axes[1])
    axes[2].scatter(F, np.unwrap(np.angle(S)), alpha=0.5, linewidths=0.0)
    axes[2].plot(F, np.unwrap(np.angle(S21_fit)) - match * 2 * π, 'r-')
    axes[2].text(0.025, 0.95, r"n = {} (DOF)".format(dof), ha="left", va="top", transform=axes[2].transAxes)
    axes[2].text(0.025, 0.87, r"$R^2 = {:.3f}$%".format(r2 * 100), ha="left", va="top", transform=axes[2].transAxes)
    axes[2].set_xlim(F[0], F[-1])
    axes[2].set_xlabel(r"Frequency (MHz)")
    axes[2].set_ylabel(r"$arg~S_{21}$")
    axes[2].grid()
    return p, perr, r2, fig



import quick.experiment as experiment
import quick.helper as helper
import numpy as np
from scipy.optimize import minimize, curve_fit
from scipy.ndimage import convolve1d, median_filter
from scipy.signal import find_peaks, peak_widths
from tqdm.notebook import tqdm
import matplotlib.pyplot as plt
import time, os

relevant_var = {
    "BaseAuto": [],
    "ReadoutLength": ["r_length"],
    "Resonator": ["r_freq", "r_power"],
    "QubitFreq": ["q_freq"],
    "PiPulseLength": ["q_length"],
    "PiPulseFreq": ["q_freq"],
    "ReadoutFreq": ["r_freq"],
    "ReadoutState": ["r_threshold", "r_phase"],
    "Relax": ["r_relax"],
    "Readout": ["r_power", "r_length"],
    "Ramsey": ["q_freq"]
}

class BaseAuto:
    var = {}
    data = None
    def __init__(self, var, silent=False, data_path=None, soccfg=None, soc=None):
        self.var = dict(var)
        self.silent = silent
        self.data_path = data_path
        self.soccfg = soccfg
        self.soc = soc
    def load_data(self, *paths):
        self.data = helper.load_data(*paths).T
    def update(self, v):
        for k in relevant_var[self.__class__.__name__]:
            v[k] = self.var[k]

class Resonator(BaseAuto):
    def calibrate(self, **kwargs):
        self.var["r_relax"] = 0
        if self.data is None:
            self.data = experiment.ResonatorSpectroscopy(
                data_path=self.data_path, title=f'(auto.Resonator) {int(self.var["r_freq"])}',
                r_power=np.arange(-50, 0, 2), r_freq=np.linspace(self.var["r_freq"] - 2, self.var["r_freq"] + 2, 101),
                p0_mode="periodic", r0_length=213, hard_avg=10, # VNA style
                soccfg=self.soccfg, soc=self.soc, var=self.var, **kwargs
            ).run(silent=self.silent).data.T
            experiment.LoopBack(soccfg=self.soccfg, soc=self.soc, var=self.var).run(silent=True)
        P, F, A = self.data[0], self.data[1], self.data[2]
        Fn = len(np.unique(F))
        F, P = F[0:Fn], P[0::Fn]
        unit = (F[-1] - F[0]) / Fn
        A = A.reshape((-1, Fn))
        avg = np.mean(A, axis=0)
        avg = convolve1d(avg, np.ones(3) / 3)
        peaks, _ = find_peaks(-avg, distance=0.3/unit, width=0.05/unit, prominence=0.3, height=-((np.max(avg) - np.min(avg)) * 0.7 + np.min(avg)))
        fig, axes = plt.subplots(1, 2, figsize=(16, 8))
        axes[0].plot(F, avg)
        axes[0].vlines(peaks * unit + F[0], ymin=np.min(avg), ymax=np.max(avg), color="red")
        axes[0].set_xlabel("Frequency (MHz)")
        axes[0].set_ylabel("Amplitude (dB)")
        axes[0].grid()
        fi = peaks[-1]
        s = convolve1d(A[:, fi], np.ones(5) / 5)
        axes[1].plot(P, s, marker="o")
        axes[1].grid()
        axes[1].set_xlabel("Power (dB)")
        if s[-1] - np.min(s) < np.std(A):
            return False, fig
        threshold = (np.max(s) - np.min(s)) * 0.4 + np.min(s)
        for i in range(len(s) - 1, 1, -1):
            if s[i] < threshold and (s[i] - s[i-1]) / (P[i] - P[i-1]) < 0.1:
                pi = i
                break
        fi = np.argmin(convolve1d(A[pi, :], np.ones(3) / 3))
        axes[1].vlines(P[pi], ymin=np.min(s), ymax=np.max(s), color="red")
        self.var["r_freq"] = float(F[fi])
        self.var["r_power"] = float(P[pi])
        return self.var, fig

class ReadoutLength(BaseAuto):
    def calibrate(self, **kwargs):
        self.var["r_relax"] = 0.5
        def scan(label, f, w):
            self.data = experiment.ResonatorSpectroscopy(
                data_path=self.data_path, title=f'(auto.ReadoutLength) {int(self.var["r_freq"])} {label}',
                r_freq=np.linspace(f - w, f + w, 1001),
                p0_mode="periodic", r0_length=213, hard_avg=100, # VNA style
                soccfg=self.soccfg, soc=self.soc, var=self.var, **kwargs
            ).run(silent=self.silent).data.T
        if self.data is None:
            scan("wide", self.var["r_freq"], 2)
        data = self.data
        p, perr, r2, fig = helper.fitResonator(data[0], data[3] + 1j * data[4], fit="circle")
        if r2 < 0.5 or perr[1] > 0.5 * p[1]:
            experiment.LoopBack(soccfg=self.soccfg, soc=self.soc, var=self.var).run(silent=True)
            return False, fig
        f, Qc = p[2], p[1]
        last_data = self.data
        scan("focus", f, f / Qc / 2)
        experiment.LoopBack(soccfg=self.soccfg, soc=self.soc, var=self.var).run(silent=True)
        data = np.hstack([last_data, self.data])
        p, perr, r2, fig = helper.fitResonator(data[0], data[3] + 1j * data[4], fit="circle")
        if r2 < 0.5 or perr[1] > 0.5 * p[1]:
            return False, fig
        f, Qc = p[2], p[1]
        self.var["r_length"] = float(min(Qc / f, 10))
        return self.var, fig

class QubitFreq(BaseAuto):
    def calibrate(self, q_freq_min=3000, q_freq_max=5000, q_gain=0.5, **kwargs):
        self.var["r_relax"] = 1
        self.var["q_gain"] = q_gain
        self.var["q_length"] = 2
        def scan(scan_freq, title=""):
            self.data = experiment.QubitSpectroscopy(data_path=self.data_path, title=f'(auto.QubitFreq) {int(self.var["r_freq"])} {title}', q_freq=scan_freq, soccfg=self.soccfg, soc=self.soc, var=self.var, **kwargs).run(silent=self.silent).data.T
        if self.data is None: 
            scan(np.arange(q_freq_min, q_freq_max, 1))
        fig, ax = plt.subplots()
        F, A = self.data[0], self.data[1]
        Fn = len(F)
        unit = (F[-1]-F[0])/Fn
        ax.scatter(F, A, color="black", label="raw data", s=20)
        std = np.std(A)
        med = np.median(A)
        A = median_filter(A, size = 5)
        ax.plot(F, A, color="red", label="median filtered")
        ax.legend()
        ax.set_xlabel("Qubit Frequency (MHz)")
        ax.set_ylabel("Amplitude [lin mag]")
        ax.set_title("QubitSpectroscopy gain=0.5")
        ax.hlines([med + 2*std, med - 2*std], xmin=F[0], xmax=F[-1], color="green")
        ax.grid()
        peak, _ = find_peaks(A, height=med+2*std, distance=10/unit)
        ax.scatter(F[peak], A[peak], marker="x", color="red")
        print("Peaks: ", F[peak])
        if len(peak) == 0:
            return False, fig
        if len(peak) == 1:
            self.var["q_freq"] = float(F[peak[0]])
            return self.var, fig
        width = peak_widths(A, peak, rel_height=0.7)
        ls = []
        for w in range(len(width[0])):
            ls += [[F[peak[w]] - width[0][w]*unit, F[peak[w]] + width[0][w]*unit]]
        for gain in np.arange(0.05, 0.4, 0.1):
            interval_score = []
            interval_peak = []
            self.var["q_gain"] = gain
            for i, l in enumerate(ls):
                scan(np.arange(l[0], l[1], 0.5), title=f"({i})gain={round(gain, 2)}")
                _A = median_filter(self.data[1], size=3)
                interval_score.append(max(np.max(_A) - med - 1.5*std, 0) / std)
                interval_peak.append(np.argmax(_A) * 0.5 + l[0])
            print("Interval Score: ", interval_score)
            _score = np.sort(interval_score)
            if _score[-1] > _score[-2] + 0.2:
                li = np.argmax(interval_score)
                print(interval_peak[li])
                self.var["q_freq"] = float(interval_peak[li])
                return self.var, fig
        return False, fig

class PiPulseLength(BaseAuto):
    def calibrate(self, q_length_max=0.5, cycles=[], tol=0.5, **kwargs):
        fig, axes = plt.subplots(len(cycles) + 1, 1)
        def scan(scan_length, cycle=0):
            self.data = experiment.Rabi(soccfg=self.soccfg, soc=self.soc, var=self.var, data_path=self.data_path, title=f"(auto.PiPulseLength) {int(self.var['r_freq'])} cycle={cycle}", q_length=scan_length, cycle=cycle, rep=2000, **kwargs).run(silent=self.silent).data.T
        def fit(cycle=0, ax=axes):
            L, A = self.data[0], self.data[2]
            def m(x, p1, p2, p3):
                return p1 * np.sin(x * p2) ** 2 + p3
            p, pcov = curve_fit(m, L, A, p0=[np.max(A) - np.min(A), helper.estimateOmega(L, A) / 2, np.min(A)], bounds=([0.1, 0.1, 0.1], [np.inf, np.inf, np.inf]))
            T = np.pi / p[1]
            residuals = A - m(L, *p)
            dof = len(L) - len(p)
            r2 = 1 - np.sum((residuals - np.mean(residuals))**2) / np.sum((A - np.mean(A))**2)
            ax.scatter(L, A, color="black", s=20)
            ax.plot(L, m(L, *p), color="blue")
            ax.set_xlabel("Pi Pulse Length (us)")
            self.var["q_length"] = float(T * (cycle / 2 + 0.5))
            ax.vlines([self.var["q_length"]], ymin=np.min(A), ymax=np.max(A), color="red")
            print("R^2 =", r2)
            return r2
        if self.data is None:
            scan(np.linspace(0.008, q_length_max, 100))
        r2 = fit(cycle=0, ax=(axes[0] if len(cycles) > 0 else axes))
        if r2 < tol:
            return False, fig
        for j in range(len(cycles)):
            c = cycles[j]
            if c <= 0:
                continue
            scan(np.linspace(self.var["q_length"] * (c - 0.5) / (c + 0.5), self.var["q_length"] * (c + 1.5) / (c + 0.5), 51), cycle=c)
            r2 = fit(cycle=c, ax=axes[j+1])
            if r2 < tol:
                return False, fig
        return self.var, fig

class PiPulseFreq(BaseAuto):
    def calibrate(self, cycles=[], **kwargs):
        fig, axes = plt.subplots(len(cycles) + 1, 1)
        r = 1 / self.var["q_length"]
        def scan(cycle=0):
            self.data = experiment.Rabi(soccfg=self.soccfg, soc=self.soc, var=self.var, data_path=self.data_path, title=f"(auto.PiPulseFreq) {int(self.var['r_freq'])} cycle={cycle}", q_freq=np.linspace(self.var["q_freq"]-r,self.var["q_freq"]+r,101), cycle=cycle, rep=2000, **kwargs).run(silent=self.silent).data.T
        def fit(ax=axes):
            F, A = self.data[0], self.data[2]
            self.var["q_freq"] = float(helper.symmetryCenter(F, A))
            ax.plot(F, A, marker="o")
            ax.set_xlabel("Pi Pulse Freq (MHz)")
            ax.vlines([self.var["q_freq"]], ymin=np.min(A), ymax=np.max(A), color="red")
        if self.data is None:
            scan(cycle=0)
        fit(ax=(axes[0] if len(cycles) > 0 else axes))
        for j in range(len(cycles)):
            c = cycles[j]
            if c <= 0:
                continue
            scan(cycle=c)
            fit(ax=axes[j+1])
        return self.var, fig

class ReadoutFreq(BaseAuto):
    def calibrate(self, r=2, **kwargs):
        if self.data is None:
            self.data = experiment.DispersiveSpectroscopy(soccfg=self.soccfg, soc=self.soc, var=self.var, data_path=self.data_path, title=f"(auto.ReadoutFreq) {int(self.var['r_freq'])}", r_freq=np.linspace(self.var["r_freq"] - r, self.var["r_freq"] + r, 201), **kwargs).run(silent=self.silent).data.T
        F, A0, P0, I0, Q0, A1, P1, I1, Q1 = self.data
        dS_amp = np.convolve(np.abs(I0 + 1j*Q0 - I1 - 1j*Q1), np.ones(8) / 8, "same")
        i0 = np.argmin(A0)
        i1 = np.argmin(A1)
        start, end = min(i0, i1), max(i0, i1)
        self.var["r_freq"] = float(F[np.argmax(dS_amp[start:end]) + start])
        fig, ax = plt.subplots()
        ax.plot(F, A0, label="Amplitude 0 (dB)", marker="o")
        ax.plot(F, A1, label="Amplitude 1 (dB)", marker="o")
        axr = ax.twinx() 
        axr.set_ylabel('Separation') 
        axr.plot(F, dS_amp, label="Separation", color="green")
        ax.vlines([self.var["r_freq"]], ymin=np.min(A0), ymax=np.max(A0), colors="red", label="Max Separation")
        ax.legend()
        ax.set_xlabel("Readout Frequency (MHz)")
        ax.set_ylabel("Amplitude (dB)")
        ax.set_title("DispersiveSpectroscopy")
        return self.var, fig

class ReadoutState(BaseAuto):
    def calibrate(self, tol=0.1, **kwargs):
        self.var["r_phase"] = 0
        if self.data is None:
            self.data = experiment.IQScatter(var=self.var, soccfg=self.soccfg, soc=self.soc, data_path=self.data_path, title=f"(auto.ReadoutState)", **kwargs).run(silent=self.silent).data.T
        self.var["r_phase"], self.var["r_threshold"], visibility, Fg, Fe, c0, c1, fig = helper.iq_scatter(self.data[0] + 1j * self.data[1], self.data[2] + 1j * self.data[3])
        if visibility < tol:
            return False, fig
        return self.var, fig 

class Ramsey(BaseAuto):
    def calibrate(self, fringe_freq=100, **kwargs):
        self.var["fringe_freq"] = fringe_freq
        max_time = 10 / fringe_freq
        if self.data is None:
            self.data = experiment.T2Ramsey(soccfg=self.soccfg, soc=self.soc, var=self.var, data_path=self.data_path, title=f"(auto.Ramsey) {int(self.var['r_freq'])}", time=np.linspace(0, max_time, 101), **kwargs).run(silent=self.silent).data.T
        L, A = self.data[0], self.data[1]
        def m(x, p1, p2, p3):
            return p1 * np.cos(p2 * x) + p3
        p, pcov = curve_fit(m, L, A, p0=[(np.max(A) - np.min(A)) / 2, helper.estimateOmega(L, A), 0.5])
        fig, ax = plt.subplots()
        ax.scatter(L, A, s=10, color="black")
        ax.plot(np.arange(L[0], L[-1], 0.001), m(np.arange(L[0], L[-1], 0.001), *p), color="red")
        ax.set_title(f"Ramsey (fringe_freq = {self.var['fringe_freq']})")
        self.var["q_freq"] = float(self.var["q_freq"] + p[1] / 2 / np.pi - self.var["fringe_freq"])
        return self.var, fig

class Readout(BaseAuto):
    def calibrate(self):
        self.var["r_length"] = self.var.get("r_length", experiment.var["r_length"])
        def negative_score(x):
            self.var["r_power"], self.var["r_length"] = float(x[0]), float(x[1])
            data = experiment.IQScatter(var=self.var, soccfg=self.soccfg, soc=self.soc).run(silent=True).data.T
            phase, threshold, visibility, Fg, Fe, c0, c1, fig = helper.iq_scatter(data[0] + 1j * data[1], data[2] + 1j * data[3], plot=False)
            c1m = np.median(data[2]) + 1j * np.median(data[3])
            c1_shift = np.abs(c1 - c1m) / np.abs(c1m - c0)
            score = visibility - abs(Fg - Fe) - 10 * c1_shift ** 2
            print(x, score)
            return -score
        initial_simplex = [[self.var["r_power"], self.var["r_length"]], [self.var["r_power"] + 4, self.var["r_length"]], [self.var["r_power"], self.var["r_length"] + 2]]
        bounds = [(-60, 0), (0.1, 10)]
        minimize(negative_score, x0=[self.var["r_power"], self.var["r_length"]], bounds=bounds, method="Nelder-Mead", options={ "maxfev": 100, "xatol": 0.05, "fatol": 1, "initial_simplex": initial_simplex })
        return self.var, None

class Relax(BaseAuto):
    def calibrate(self, **kwargs):
        if self.data is None:
            self.data = experiment.T1(var=self.var, data_path=self.data_path, soccfg=self.soccfg, soc=self.soc, title=f"(auto.Relax) {int(self.var['r_freq'])}", time=np.linspace(0, self.var["r_relax"] * 0.8, 61), **kwargs).run(silent=self.silent).data.T
        p, _, _, fig = helper.fitT1(self.data[0], self.data[1])
        print("T1 =", p[1])
        self.var["r_relax"] = float(5 * p[1])
        return self.var, fig

def run(path, soccfg=None, soc=None, data_path=None):
    config = helper.load_yaml(path)
    qubits = config["qubits"]
    steps = config["steps"]
    qi = -1 # find the qubit to be run
    min_run = 9e9
    for i, q in enumerate(qubits):
        if q["status"].get("step", "start") in ["end", "fail", "pause"]:
            continue  # completed
        run = q["status"].get("run", 0)
        if run < min_run:
            qi = i
            min_run = run
    config["current"] = qi
    config["time"] = int(time.time() * 1000)
    helper.save_yaml(path, config)
    if qi < 0: # all completed
        return False
    step = qubits[qi]["status"].get("step", "start")
    print(f"\n------------ quick.auto.run [{qi}] {step} ------------\n")
    skip = False
    try:
        a = globals()[steps[step].get("class", step)](var=qubits[qi]["var"], soccfg=soccfg, soc=soc, data_path=data_path)
    except:
        skip = True
        v = True
    try:
        if skip is False:
            v, fig = a.calibrate(**steps[step].get("argument", {}), **qubits[qi]["argument"].get(step, {}))
    except KeyboardInterrupt:
        print("\n!!!!! KeyboardInterrupt !!!!!")
        _config = helper.load_yaml(path) # avoid overwrite
        _config["current"] = -2
        _config["time"] = int(time.time() * 1000)
        helper.save_yaml(path, _config)
        return
    except Exception as e:
        print(e)
        v = False
    qubits[qi]["status"]["run"] = qubits[qi]["status"].get("run", 0) + 1
    qubits[qi]["status"][step] = qubits[qi]["status"].get(step, 0) + 1
    i = qubits[qi]["status"][step]
    if v is False: # failed
        qubits[qi]["status"]["step"] = steps[step].get(f"back{i}", steps[step].get("back", step))
    else: # succeeded
        if skip is False:
            a.update(qubits[qi]["var"])
        qubits[qi]["status"]["step"] = steps[step].get(f"next{i}", steps[step].get("next", "end"))
    _config = helper.load_yaml(path) # avoid overwrite
    _config["current"] = -2
    _config["time"] = int(time.time() * 1000)
    _config["qubits"][qi] = qubits[qi]
    helper.save_yaml(path, _config)
    return True

